'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import gevent
from gevent.coros import Semaphore
from gevent import socket
from gevent.pool import Pool

from haigha.transports import Transport

class TransportTest(Chai):

  def test_init_and_connection_property(self):
    t = Transport( 'conn' )
    assert_equals( 'conn', t._connection )
    assert_equals( 'conn', t.connection )

  def test_process_channels(self):
    t = Transport('conn')
    ch1 = mock()
    ch2 = mock()
    chs = set([ch1,ch2])
    expect( ch1.process_frames )
    expect( ch2.process_frames )

    t.process_channels( chs )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

class ChannelPool(object):
  '''
  Manages a pool of channels for transaction-based publishing.  This allows a
  client to use as many channels as are necessary to publish while not creating
  a backlog of transactions that slows throughput and consumes memory.

  The pool can accept an optional `size` argument in the ctor, which caps the
  number of channels which the pool will allocate. If no channels are available
  on `publish()`, the message will be locally queued and sent as soon as a 
  channel is available. It is recommended that you use the pool with a max
  size, as each channel consumes memory on the broker and it is possible to
  exercise memory limit protection seems on the broker due to number of
  channels.
  '''
  
  def __init__(self, connection, size=None):
    '''Initialize the channel on a connection.'''
    self._connection = connection
    self._free_channels = set()
    self._size = size
    self._queue = deque()
    self._channels = 0

  def publish(self, *args, **kwargs):
    '''
    Publish a message. Caller can supply an optional callback which will
    be fired when the transaction is committed. Tries very hard to avoid
    closed and inactive channels, but a ChannelError or ConnectionError
    may still be raised.
    '''
    user_cb = kwargs.pop('cb', None)

    # If the first channel we grab is inactive, continue fetching until
    # we get an active channel, then put the inactive channels back in
    # the pool. Try to keep the overhead to a minimum.
    channel = self._get_channel()

    if channel and not channel.active:
      inactive_channels = set()
      while channel and not channel.active:
        inactive_channels.add( channel )
        channel = self._get_channel()
      self._free_channels.update( inactive_channels )

    # When the transaction is committed, add the channel back to the pool and
    # call any user-defined callbacks. If there is anything in queue, pop it
    # and call back to publish(). Only do so if the channel is still active
    # though, because otherwise the message will end up at the back of the
    # queue, breaking the original order.
    def committed():
      self._free_channels.add( channel )
      if channel.active and not channel.closed:
        self._process_queue()
      if user_cb is not None: user_cb()

    if channel:
      channel.publish_synchronous( *args, cb=committed, **kwargs )
    else:
      kwargs['cb'] = user_cb
      self._queue.append( (args,kwargs) )
    
  def _process_queue(self):
    '''
    If there are any message in the queue, process one of them.
    '''
    if len(self._queue):
      args, kwargs = self._queue.popleft()
      self.publish( *args, **kwargs )

  def _get_channel(self):
    '''
    Fetch a channel from the pool. Will return a new one if necessary. If
    a channel in the free pool is closed, will remove it. Will return None
    if we hit the cap. Will clean up any channels that were published to but
    closed due to error.
    '''
    while len(self._free_channels):
      rval = self._free_channels.pop()
      if not rval.closed: 
        return rval
      # don't adjust _channels value because the callback will do that and
      # we don't want to double count it.

    if not self._size or self._channels < self._size:
      rval = self._connection.channel()
      self._channels += 1
      rval.add_close_listener( self._channel_closed_cb )
      return rval

  def _channel_closed_cb(self, channel):
    '''
    Callback when channel closes.
    '''
    self._channels -= 1
    self._process_queue()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames import MethodFrame
from haigha.classes import ProtocolClass

from collections import deque

class TransactionClass(ProtocolClass):
  '''
  Implements the AMQP Transaction class
  '''

  class TransactionsNotEnabled(ProtocolClass.ProtocolError): 
    '''Tried to use transactions without enabling them.'''

  def __init__(self, *args, **kwargs):
    super(TransactionClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_select_ok,
      21 : self._recv_commit_ok,
      31 : self._recv_rollback_ok,
    }

    self._enabled = False
    self._select_cb = deque()
    self._commit_cb = deque()
    self._rollback_cb = deque()

  @property
  def enabled(self):
    '''Get whether transactions have been enabled.'''
    return self._enabled
  
  def _cleanup(self):
    '''
    Cleanup all the local data.
    '''
    self._select_cb = None
    self._commit_cb = None
    self._rollback_cb = None
    super(TransactionClass,self)._cleanup()
  
  def select(self, cb=None):
    '''
    Set this channel to use transactions.
    '''
    if not self._enabled:
      self._enabled = True
      self.send_frame( MethodFrame(self.channel_id, 90, 10) )
      self.channel.add_synchronous_cb( self._recv_select_ok )
      self._select_cb.append(cb)

  def _recv_select_ok(self, _method_frame):
    cb = self._select_cb.popleft()
    if cb: cb()
    
  def commit(self, cb=None):
    '''
    Commit the current transaction.  Caller can specify a callback to use
    when the transaction is committed.
    '''
    # Could call select() but spec 1.9.2.3 says to raise an exception
    if not self.enabled: raise self.TransactionsNotEnabled()

    self.send_frame( MethodFrame(self.channel_id, 90, 20) )
    self.channel.add_synchronous_cb( self._recv_commit_ok )
    self._commit_cb.append( cb )

  def _recv_commit_ok(self, _method_frame):
    cb = self._commit_cb.popleft()
    if cb: cb()

  def rollback(self, cb=None):
    '''
    Abandon all message publications and acks in the current transaction.
    Caller can specify a callback to use when the transaction has been
    aborted.
    '''
    # Could call select() but spec 1.9.2.5 says to raise an exception
    if not self.enabled: raise self.TransactionsNotEnabled()

    self.send_frame( MethodFrame(self.channel_id, 90, 30) )
    self.channel.add_synchronous_cb( self._recv_rollback_ok )
    self._rollback_cb.append( cb )

  def _recv_rollback_ok(self, _method_frame):
    cb = self._rollback_cb.popleft()
    if cb: cb()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
from collections import deque

from haigha.channel import Channel
from haigha.exceptions import ChannelError, ChannelClosed
from haigha.classes import *
from haigha.frames import *

class ChannelTest(Chai):

  def test_init(self):
    c = Channel('connection', 'id')
    self.assertEquals( 'connection', c._connection )
    self.assertEquals( 'id', c._channel_id )
    self.assertTrue( isinstance(c.channel, ChannelClass) )
    self.assertTrue( isinstance(c.exchange, ExchangeClass) )
    self.assertTrue( isinstance(c.queue, QueueClass) )
    self.assertTrue( isinstance(c.basic, BasicClass) )
    self.assertTrue( isinstance(c.tx, TransactionClass) )
    self.assertEquals( c._class_map[20], c.channel )
    self.assertEquals( c._class_map[40], c.exchange )
    self.assertEquals( c._class_map[50], c.queue )
    self.assertEquals( c._class_map[60], c.basic )
    self.assertEquals( c._class_map[90], c.tx )
    self.assertEquals( deque([]), c._pending_events )
    self.assertEquals( deque([]), c._frame_buffer )
    assert_equals( set([]), c._close_listeners )

  def test_properties(self):
    connection = mock()
    connection.logger = 'logger'

    c = Channel(connection, 'id')
    c.channel = mock()
    c.channel.closed = 'closed'
    c.channel.close_info = 'uwerebad'

    assertEquals( connection, c.connection )
    assertEquals( 'id', c.channel_id )
    assertEquals( 'logger', c.logger )
    assertEquals( 'closed', c.closed )
    assertEquals( 'uwerebad', c.close_info )

  def test_add_close_listener(self):
    c = Channel(None,None)
    c.add_close_listener('foo')
    assert_equals( set(['foo']), c._close_listeners )

  def test_remove_close_listener(self):
    c = Channel(None,None)
    c.add_close_listener('foo')
    c.remove_close_listener('foo')
    c.remove_close_listener('bar')
    assert_equals( set([]), c._close_listeners )

  def test_open(self):
    c = Channel(None,None)
    expect( c.channel.open )
    c.open()
 
  def test_active(self):
    c = Channel(None, None)
    expect( c.channel.open )
    c.open()
    assertTrue(c.active)

  def test_close_with_no_args(self):
    c = Channel(None,None)
    expect( c.channel.close ).args( 0, '', 0, 0 )
    c.close()
  
  def test_close_with_args(self):
    c = Channel(None,None)
    expect( c.channel.close ).args(1, 'two', 3, 4)
    expect( c.channel.close ).args(1, 'two', 3, 4)

    c.close(1, 'two', 3, 4)
    c.close(reply_code=1, reply_text='two', class_id=3, method_id=4)

  def test_publish(self):
    c = Channel(None,None)
    expect( c.basic.publish ).args( 'arg1', 'arg2', foo='bar' )
    c.publish( 'arg1', 'arg2', foo='bar' )

  def test_publish_synchronous(self):
    c = Channel(None,None)
    expect( c.tx.select )
    expect( c.basic.publish ).args( 'arg1', 'arg2', foo='bar' )
    expect( c.tx.commit ).args( cb='a_cb' )

    c.publish_synchronous( 'arg1', 'arg2', foo='bar', cb='a_cb' )

  def test_dispatch(self):
    c = Channel(None,None)
    frame = mock()
    frame.class_id = 32
    klass = mock()
    
    c._class_map[32] = klass
    expect( klass.dispatch ).args( frame )
    c.dispatch( frame )

    frame.class_id = 33
    assert_raises( Channel.InvalidClass, c.dispatch, frame )

  def test_buffer_frame(self):
    c = Channel(None,None)
    c.buffer_frame( 'f1' )
    c.buffer_frame( 'f2' )
    assertEquals( deque(['f1', 'f2']), c._frame_buffer )

  def test_process_frames_with_one_frame(self):
    conn = mock()
    conn.logger = mock()
    c = Channel(conn, None)
    ch_id, c_id, m_id = 0, 20, 2
    f = MethodFrame(ch_id, c_id, m_id)
    c._frame_buffer = deque([ f ])
    expect(c.logger.error).args(ignore(), ignore(), exc_info=ignore())
    expect(c.close).args(ignore(), ignore())

    c.process_frames()

  def test_process_frames_with_two_frames(self):
    conn = mock()
    conn.logger = mock()
    c = Channel(conn, None)
    ch_id, c_id, m_id = 0, 1, 2
    f0 = MethodFrame(ch_id, c_id, m_id)
    f1 = MethodFrame(ch_id, c_id, m_id)
    c._frame_buffer = deque([ f0, f1 ])
    expect(c.logger.error).args(ignore(), ignore(), exc_info=ignore())
    expect(c.close).args(ignore(), ignore())

    c.process_frames()

  def test_next_frame_with_a_frame(self):
    c = Channel(None, None)
    ch_id, c_id, m_id = 0, 1, 2
    f0 = MethodFrame(ch_id, c_id, m_id)
    f1 = MethodFrame(ch_id, c_id, m_id)
    c._frame_buffer = deque([ f0, f1 ])
    assertEquals(c.next_frame(), f0)

  def test_next_frame_with_no_frames(self):
    c = Channel(None, None)
    c._frame_buffer = deque()
    assertEquals(c.next_frame(), None)

  def test_requeue_frames(self):
    c = Channel(None, None)
    ch_id, c_id, m_id = 0, 1, 2
    f = [MethodFrame(ch_id, c_id, m_id) for i in xrange(4)]
    c._frame_buffer = deque(f[:2])

    c.requeue_frames(f[2:])
    assertEquals(c._frame_buffer, deque([f[i] for i in [3, 2, 0, 1]]))

  def test_process_frames_when_no_frames(self):
    # Not that this should ever happen, but to be sure
    c = Channel(None,None)
    stub( c.dispatch )
    c.process_frames()

  def test_process_frames_stops_when_buffer_is_empty(self):
    c = Channel(None, None)
    f0 = MethodFrame('ch_id', 'c_id', 'm_id')
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 )
    expect( c.dispatch ).args( f1 )

    c.process_frames()

  def test_process_frames_stops_when_frameunderflow_raised(self):
    c = Channel(None, None)
    f0 = MethodFrame('ch_id', 'c_id', 'm_id')
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 ).raises( ProtocolClass.FrameUnderflow )

    c.process_frames()
    assertEquals( f1, c._frame_buffer[0] )
  
  def test_process_frames_logs_and_closes_when_dispatch_error_raised(self):
    c = Channel(None, None)
    c._connection = mock()
    c._connection.logger = mock()
    
    f0 = MethodFrame(20, 30, 40)
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 ).raises( Exception("zomg it broked") )
    expect( c._connection.logger.error ).args( "Failed to dispatch %s", f0, exc_info=True )
    expect( c.close ).args( 500, str )

    c.process_frames()
  
  def test_send_frame_when_not_closed_no_flow_control_no_pending_events(self):
    conn = mock()
    c = Channel(conn, 32)

    expect( conn.send_frame ).args('frame')
    
    c.send_frame( 'frame' )

  def test_send_frame_when_not_closed_no_flow_control_pending_event(self):
    conn = mock()
    c = Channel(conn, 32)
    c._pending_events.append( 'cb' )

    c.send_frame( 'frame' )
    self.assertEquals( deque(['cb','frame']), c._pending_events )

  def test_send_frame_when_not_closed_and_flow_control(self):
    conn = mock()
    c = Channel(conn, 32)
    c.channel._active = False

    method = MethodFrame(1,2,3)
    heartbeat = HeartbeatFrame()
    header = HeaderFrame(1,2,3,4)
    content = ContentFrame(1,'foo')
    
    expect( conn.send_frame ).args( method )
    expect( conn.send_frame ).args( heartbeat )

    c.send_frame( method )
    c.send_frame( heartbeat )
    self.assertRaises( Channel.Inactive, c.send_frame, header )
    self.assertRaises( Channel.Inactive, c.send_frame, content )

  def test_send_frame_when_closed_for_a_reason(self):
    conn = mock()
    c = Channel(conn, 32)
    c.channel._closed = True
    c.channel._close_info = {'reply_code':42, 'reply_text':'bad'}

    assertRaises( ChannelClosed, c.send_frame, 'frame' )
  
  def test_send_frame_when_closed_for_no_reason(self):
    conn = mock()
    c = Channel(conn, 32)
    c.channel._closed = True
    c.channel._close_info = {'reply_code':42, 'reply_text':''}

    assertRaises( ChannelClosed, c.send_frame, 'frame' )

  def test_add_synchronous_cb(self):
    c = Channel(None,None)

    assertEquals( deque([]), c._pending_events )
    c.add_synchronous_cb( 'foo' )
    assertEquals( deque(['foo']), c._pending_events )

  def test_clear_synchronous_cb_when_no_pending(self):
    c = Channel(None,None)
    stub( c._flush_pending_events )

    assertEquals( deque([]), c._pending_events )
    c.clear_synchronous_cb( 'foo' )

  def test_clear_synchronous_cb_when_pending_cb_matches(self):
    c = Channel(None,None)
    c._pending_events = deque(['foo'])

    expect( c._flush_pending_events )

    c.clear_synchronous_cb( 'foo' )
    assertEquals( deque([]), c._pending_events )

  def test_clear_synchronous_cb_when_pending_cb_doesnt_match_but_isnt_in_list(self):
    c = Channel(None,None)
    c._pending_events = deque(['foo'])

    expect( c._flush_pending_events )

    c.clear_synchronous_cb( 'bar' )
    assertEquals( deque(['foo']), c._pending_events )

  def test_clear_synchronous_cb_when_pending_cb_doesnt_match_but_isnt_in_list(self):
    c = Channel(None,None)
    stub( c._flush_pending_events )
    c._pending_events = deque(['foo', 'bar'])

    assertRaises( ChannelError, c.clear_synchronous_cb, 'bar' )
    assertEquals( deque(['foo','bar']), c._pending_events )

  def test_flush_pending_events_flushes_all_leading_frames(self):
    conn = mock()
    c = Channel(conn,42)
    f1 = MethodFrame(1,2,3)
    f2 = MethodFrame(1,2,3)
    f3 = MethodFrame(1,2,3)
    c._pending_events = deque([f1, f2, 'cb', f3])

    expect( conn.send_frame ).args( f1 )
    expect( conn.send_frame ).args( f2 )

    c._flush_pending_events()
    assertEquals( deque(['cb',f3]), c._pending_events )

  def test_closed_cb_without_final_frame(self):
    c = Channel('connection',None)
    c._pending_events = 'foo'
    c._frame_buffer = 'foo'
    listener = mock()
    c._close_listeners = set([listener])
    
    for val in c._class_map.values():
      expect( val._cleanup )
    expect( listener ).args( c )

    c._closed_cb()
    assert_equals( deque([]), c._pending_events )
    assert_equals( deque([]), c._frame_buffer )
    assert_equals( None, c._connection )
    assert_equals( None, c.channel )
    assert_equals( None, c.exchange )
    assert_equals( None, c.queue )
    assert_equals( None, c.basic )
    assert_equals( None, c.tx )
    assert_equals( None, c._class_map )
    assert_equals( set(), c._close_listeners )

  def test_closed_cb_with_final_frame(self):
    conn = mock()
    c = Channel(conn,None)

    expect( conn.send_frame ).args('final')    
    for val in c._class_map.values():
      expect( val._cleanup )

    c._closed_cb('final')



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from struct import unpack, unpack_from, Struct
from datetime import datetime
from decimal import Decimal

class Reader(object):
  """
  A stream-like reader object that supports all the basic data types of AMQP.
  """

  class ReaderError(Exception): '''Base class for all reader errors.'''
  class BufferUnderflow(ReaderError): '''Not enough bytes to satisfy the request.'''
  class FieldError(ReaderError): '''Unsupported field type was read.'''

  def __init__(self, source, start_pos=0, size=None):
    """
    source should be a bytearray, io object with a read() method, another
    Reader, a plain or unicode string. Can be allocated over a slice of source.
    """
    # Note: buffer used here because unpack_from can't accept an array,
    # which I think is related to http://bugs.python.org/issue7827
    if isinstance(source, bytearray):
      self._input = buffer(source)
    elif isinstance(source, Reader):
      self._input = source._input
    elif hasattr(source, 'read'):
      self._input = buffer( source.read() )
    elif isinstance(source, str):
      self._input = buffer(source)
    elif isinstance(source, unicode):
      self._input = buffer(source.encode('utf8'))
    else:
      raise ValueError('Reader needs a bytearray, io object or plain string')

    self._start_pos = self._pos = start_pos
    self._end_pos = len(self._input)
    if size:
      self._end_pos = self._start_pos + size

  def __str__(self):
    return ''.join( ['\\x%s'%(c.encode('hex')) for c in self._input[self._start_pos:self._end_pos]] )

  def tell(self):
    '''
    Current position
    '''
    return self._pos

  def seek(self, offset, whence=0):
    '''
    Simple seek. Follows standard interface.
    '''
    if whence==0:
      self._pos = self._start_pos + offset
    elif whence==1:
      self._pos += offset
    else:
      self._pos = ( self._end_pos-1 ) + offset

  def _check_underflow(self, n):
    '''
    Raise BufferUnderflow if there's not enough bytes to satisfy the request.
    '''
    if self._pos+n > self._end_pos:
      raise self.BufferUnderflow()

  def __len__(self):
    '''
    Supports content framing in Channel
    '''
    return self._end_pos - self._start_pos

  def buffer(self):
    '''
    Get a copy of the buffer that this is reading from. Returns a buffer object
    '''
    return buffer( self._input, self._start_pos, (self._end_pos-self._start_pos) )

  def read(self, n):
    """
    Read n bytes.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    """
    self._check_underflow( n )
    rval = self._input[ self._pos:self._pos+n ]
    self._pos += n
    return rval

  def read_bit(self):
    """
    Read a single boolean value, returns 0 or 1. Convience for single bit fields

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    """
    # Perform a faster check on underflow
    if self._pos >= self._end_pos: raise self.BufferUnderflow()
    result = ord(self._input[ self._pos ]) & 1
    self._pos += 1
    return result

  def read_bits(self, num):
    '''
    Read several bits packed into the same field. Will return as a list. The
    bit field itself is little-endian, though the order of the returned array
    looks big-endian for ease of decomposition.

    Reader('\x02').read_bits(2) -> [False,True]
    Reader('\x08').read_bits(2) -> [False,True,False,False,False,False,False,False]
    first_field, second_field = Reader('\x02').read_bits(2)

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise ValueError if num < 0 or num > 9
    '''
    # Perform a faster check on underflow
    if self._pos >= self._end_pos: raise self.BufferUnderflow()
    if num < 0 or num >= 9: raise ValueError("8 bits per field")
    field = ord(self._input[self._pos])
    result = map(lambda x: field>>x & 1, xrange(num) )
    self._pos += 1
    return result

  def read_octet(self, unpacker=Struct('B').unpack_from, size=Struct('B').size):
    """
    Read one byte, return as an integer

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    # Technically should look at unpacker.size, but skipping that is way
    # faster and this method is the most-called of the readers
    if self._pos >= self._end_pos: raise self.BufferUnderflow()
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def read_short(self, unpacker=Struct('>H').unpack_from, size=Struct('>H').size):
    """
    Read an unsigned 16-bit integer

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    self._check_underflow( size )
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def read_long(self, unpacker=Struct('>I').unpack_from, size=Struct('>I').size):
    """
    Read an unsigned 32-bit integer

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    self._check_underflow( size )
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def read_longlong(self, unpacker=Struct('>Q').unpack_from, size=Struct('>Q').size):
    """
    Read an unsigned 64-bit integer

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    self._check_underflow( size )
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def read_shortstr(self):
    """
    Read a utf-8 encoded string that's stored in up to
    255 bytes.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise UnicodeDecodeError if the text is mal-formed.
    Will raise struct.error if the data is malformed
    """
    slen = self.read_octet()
    return self.read(slen)

  def read_longstr(self):
    """
    Read a string that's up to 2**32 bytes, the encoding
    isn't specified in the AMQP spec, so just return it as
    a plain Python string.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    slen = self.read_long()
    return self.read(slen)
  
  def read_timestamp(self):
    """
    Read and AMQP timestamp, which is a 64-bit integer representing
    seconds since the Unix epoch in 1-second resolution.  Return as
    a Python datetime.datetime object, expressed as localtime.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    return datetime.fromtimestamp( self.read_longlong() )

  def read_table(self):
    """
    Read an AMQP table, and return as a Python dictionary.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise UnicodeDecodeError if the text is mal-formed.
    Will raise struct.error if the data is malformed
    """
    # Only need to check underflow on the table once
    tlen = self.read_long()
    self._check_underflow(tlen)
    end_pos = self._pos + tlen
    result = {}
    while self._pos < end_pos:
      name = self._field_shortstr()
      result[name] = self._read_field()
    return result

  def _read_field(self):
    '''
    Read a single byte for field type, then read the value.
    '''
    ftype = self._input[ self._pos ]
    self._pos += 1
    
    reader = self.field_type_map.get( ftype )
    if reader:
      return reader(self)

    raise Reader.FieldError('Unknown field type %s', ftype)
    
  def _field_bool(self):
    result = ord(self._input[ self._pos ]) & 1
    self._pos += 1
    return result

  def _field_short_short_int(self, unpacker=Struct('b').unpack_from, size=Struct('b').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval
    
  def _field_short_short_uint(self, unpacker=Struct('B').unpack_from, size=Struct('B').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def _field_short_int(self, unpacker=Struct('>h').unpack_from, size=Struct('>h').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval
    
  def _field_short_uint(self, unpacker=Struct('>H').unpack_from, size=Struct('>H').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def _field_long_int(self, unpacker=Struct('>i').unpack_from, size=Struct('>i').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval
    
  def _field_long_uint(self, unpacker=Struct('>I').unpack_from, size=Struct('>I').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def _field_long_long_int(self, unpacker=Struct('>q').unpack_from, size=Struct('>q').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval
    
  def _field_long_long_uint(self, unpacker=Struct('>Q').unpack_from, size=Struct('>Q').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  def _field_float(self, unpacker=Struct('>f').unpack_from, size=Struct('>f').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval
    
  def _field_double(self, unpacker=Struct('>d').unpack_from, size=Struct('>d').size):
    rval = unpacker( self._input, self._pos )[0]
    self._pos += size
    return rval

  # Coding to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 which
  # differs from spec in that the value is signed.
  def _field_decimal(self):
    d = self._field_short_short_uint()
    n = self._field_long_int()
    return Decimal(n) / Decimal(10 ** d)

  def _field_shortstr(self):
    slen = self._field_short_short_uint()
    rval = self._input[ self._pos:self._pos+slen ]
    self._pos += slen
    return rval
  
  def _field_longstr(self):
    slen = self._field_long_uint()
    rval = self._input[ self._pos:self._pos+slen ]
    self._pos += slen
    return rval

  def _field_array(self):
    alen = self.read_long()
    end_pos = self._pos + alen
    rval = []
    while self._pos < end_pos:
      rval.append( self._read_field() )
    return rval
     
  def _field_timestamp(self):
    """
    Read and AMQP timestamp, which is a 64-bit integer representing
    seconds since the Unix epoch in 1-second resolution.  Return as
    a Python datetime.datetime object, expressed as localtime.

    Will raise BufferUnderflow if there's not enough bytes in the buffer.
    Will raise struct.error if the data is malformed
    """
    return datetime.fromtimestamp( self._field_long_long_uint() )

  def _field_bytearray(self):
    slen = self._field_long_uint()
    rval = bytearray( self._input[ self._pos:self._pos+slen ] )
    self._pos += slen
    return rval

  def _field_none(self):
    return None

  # A mapping for quick lookups
  # Rabbit and Qpid 0.9.1 mapping
  field_type_map = {
    't' : _field_bool,
    'b' : _field_short_short_int,
    's' : _field_short_int,
    'I' : _field_long_int,
    'l' : _field_long_long_int,
    'f' : _field_float,
    'd' : _field_double,
    'D' : _field_decimal,
    'S' : _field_longstr,
    'T' : _field_timestamp,
    'F' : read_table,
    'V' : _field_none,
    'x' : _field_bytearray,
  }

  # 0.9.1 spec mapping
  #field_type_map = {
  #  't' : _field_bool,
  #  'b' : _field_short_short_int,
  #  'B' : _field_short_short_uint,
  #  'U' : _field_short_int,
  #  'u' : _field_short_uint,
  #  'I' : _field_long_int,
  #  'i' : _field_long_uint,
  #  'L' : _field_long_long_int,
  #  'l' : _field_long_long_uint,
  #  'f' : _field_float,
  #  'd' : _field_double,
  #  'D' : _field_decimal,
  #  's' : _field_shortstr,
  #  'S' : _field_longstr,
  #  'A' : _field_array,
  #  'T' : _field_timestamp,
  #  'F' : read_table,
  #  'V' : _field_none,
  #}



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import struct
import time
from datetime import datetime

from haigha.frames import header_frame
from haigha.frames.header_frame import HeaderFrame
from haigha.reader import Reader
from haigha.writer import Writer

class HeaderFrameTest(Chai):

  def test_type(self):
    assert_equals( 2, HeaderFrame.type() )

  def test_properties(self):
    frame = HeaderFrame(42, 'class_id', 'weight', 'size', 'props')
    assert_equals( 42, frame.channel_id )
    assert_equals( 'class_id', frame.class_id )
    assert_equals( 'weight', frame.weight )
    assert_equals( 'size', frame.size )
    assert_equals( 'props', frame.properties )

  def test_str(self):
    # Don't bother checking the copy
    frame = HeaderFrame(42, 5, 6, 7, 'props')
    assert_equals('HeaderFrame[channel: 42, class_id: 5, weight: 6, size: 7, properties: props]',
      str(frame) )

  def test_parse_fast_for_standard_properties(self):
    bit_writer = Writer()
    val_writer = Writer()

    # strip ms because amqp doesn't include it
    now = datetime.fromtimestamp( long(time.mktime(datetime.now().timetuple())) )

    bit_field = 0
    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      bit_field = (bit_field << 1) | 1

      if ptype=='shortstr':
        val_writer.write_shortstr( pname )
      elif ptype=='octet':
        val_writer.write_octet( 42 )
      elif ptype=='timestamp':
        val_writer.write_timestamp( now )
      elif ptype=='table':
        val_writer.write_table( {'foo':'bar'} )

    bit_field <<= (16- len(HeaderFrame.PROPERTIES))
    bit_writer.write_short( bit_field )
    
    header_writer = Writer()
    header_writer.write_short( 5 )
    header_writer.write_short( 6 )
    header_writer.write_longlong( 7 )
    payload = header_writer.buffer()
    payload += bit_writer.buffer()
    payload += val_writer.buffer()

    reader = Reader(payload)
    frame = HeaderFrame.parse(4, reader)

    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      if ptype=='shortstr':
        self.assertEquals( pname, frame.properties[pname] )
      elif ptype=='octet':
        self.assertEquals( 42, frame.properties[pname] )
      elif ptype=='timestamp':
        self.assertEquals( now, frame.properties[pname] )
      elif ptype=='table':
        self.assertEquals( {'foo':'bar'}, frame.properties[pname] )

    assert_equals( 4, frame.channel_id )
    assert_equals( 5, frame._class_id )
    assert_equals( 6, frame._weight )
    assert_equals( 7, frame._size )
  
  def test_parse_slow_for_standard_properties(self):
    HeaderFrame.DEFAULT_PROPERTIES = False
    bit_writer = Writer()
    val_writer = Writer()

    # strip ms because amqp doesn't include it
    now = datetime.fromtimestamp( long(time.mktime(datetime.now().timetuple())) )

    bit_field = 0
    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      bit_field = (bit_field << 1) | 1

      if ptype=='shortstr':
        val_writer.write_shortstr( pname )
      elif ptype=='octet':
        val_writer.write_octet( 42 )
      elif ptype=='timestamp':
        val_writer.write_timestamp( now )
      elif ptype=='table':
        val_writer.write_table( {'foo':'bar'} )

    bit_field <<= (16- len(HeaderFrame.PROPERTIES))
    bit_writer.write_short( bit_field )
    
    header_writer = Writer()
    header_writer.write_short( 5 )
    header_writer.write_short( 6 )
    header_writer.write_longlong( 7 )
    payload = header_writer.buffer()
    payload += bit_writer.buffer()
    payload += val_writer.buffer()

    reader = Reader(payload)
    frame = HeaderFrame.parse(4, reader)
    HeaderFrame.DEFAULT_PROPERTIES = True

    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      if ptype=='shortstr':
        self.assertEquals( pname, frame.properties[pname] )
      elif ptype=='octet':
        self.assertEquals( 42, frame.properties[pname] )
      elif ptype=='timestamp':
        self.assertEquals( now, frame.properties[pname] )
      elif ptype=='table':
        self.assertEquals( {'foo':'bar'}, frame.properties[pname] )

    assert_equals( 4, frame.channel_id )
    assert_equals( 5, frame._class_id )
    assert_equals( 6, frame._weight )
    assert_equals( 7, frame._size )

  def test_write_frame_fast_for_standard_properties(self):
    bit_field = 0
    properties = {}
    now = datetime.fromtimestamp( long(time.mktime(datetime.now().timetuple())) )
    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      bit_field |= mask

      if ptype=='shortstr':
        properties[pname] = pname
      elif ptype=='octet':
        properties[pname] = 42
      elif ptype=='timestamp':
        properties[pname] = now
      elif ptype=='table':
        properties[pname] = {'foo':'bar'}

    frame = HeaderFrame(42, 5,6,7, properties)
    buf = bytearray()
    frame.write_frame(buf)

    reader = Reader(buf)
    assert_equals( 2, reader.read_octet() )
    assert_equals( 42, reader.read_short() )
    size = reader.read_long()
    start_pos = reader.tell()
    assert_equals( 5, reader.read_short() )
    assert_equals( 6, reader.read_short() )
    assert_equals( 7, reader.read_longlong() )
    assert_equals( 0b1111111111111100, reader.read_short() )

    for pname, ptype, rfunc, wfunc, mask in HeaderFrame.PROPERTIES:
      if ptype=='shortstr':
        assertEquals( pname, reader.read_shortstr() )
      elif ptype=='octet':
        assertEquals( 42, reader.read_octet() )
      elif ptype=='timestamp':
        assertEquals( now, reader.read_timestamp() )
      elif ptype=='table':
        assertEquals( {'foo':'bar'}, reader.read_table() )

    end_pos = reader.tell()
    assert_equals( size, end_pos-start_pos )
    assert_equals( 0xce, reader.read_octet() )
  
  def test_write_frame_slow_for_standard_properties(self):
    HeaderFrame.DEFAULT_PROPERTIES = False
    bit_field = 0
    properties = {}
    now = datetime.fromtimestamp( long(time.mktime(datetime.now().timetuple())) )
    for pname, ptype, reader, writer, mask in HeaderFrame.PROPERTIES:
      bit_field |= mask

      if ptype=='shortstr':
        properties[pname] = pname
      elif ptype=='octet':
        properties[pname] = 42
      elif ptype=='timestamp':
        properties[pname] = now
      elif ptype=='table':
        properties[pname] = {'foo':'bar'}

    frame = HeaderFrame(42, 5,6,7, properties)
    buf = bytearray()
    frame.write_frame(buf)
    HeaderFrame.DEFAULT_PROPERTIES = True

    reader = Reader(buf)
    assert_equals( 2, reader.read_octet() )
    assert_equals( 42, reader.read_short() )
    size = reader.read_long()
    start_pos = reader.tell()
    assert_equals( 5, reader.read_short() )
    assert_equals( 6, reader.read_short() )
    assert_equals( 7, reader.read_longlong() )
    assert_equals( 0b1111111111111100, reader.read_short() )

    for pname, ptype, rfunc, wfunc, mask in HeaderFrame.PROPERTIES:
      if ptype=='shortstr':
        assertEquals( pname, reader.read_shortstr() )
      elif ptype=='octet':
        assertEquals( 42, reader.read_octet() )
      elif ptype=='timestamp':
        assertEquals( now, reader.read_timestamp() )
      elif ptype=='table':
        assertEquals( {'foo':'bar'}, reader.read_table() )

    end_pos = reader.tell()
    assert_equals( size, end_pos-start_pos )
    assert_equals( 0xce, reader.read_octet() )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.classes import ProtocolClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

class ChannelClass(ProtocolClass):
  '''
  Implements the AMQP Channel class
  '''
  
  def __init__(self, *args, **kwargs):
    super(ChannelClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_open_ok,
      20 : self._recv_flow,
      21 : self._recv_flow_ok,
      40 : self._recv_close,
      41 : self._recv_close_ok,
    }

    self._closed = False
    self._close_info = {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    }

    self._active = True
    self._flow_control_cb = None

  @property
  def closed(self):
    '''Return whether this channel has been closed.'''
    return self._closed

  @property
  def close_info(self):
    '''Return dict with information on why this channel is closed.  Will
    return None if the channel is open.'''
    return self._close_info if self._closed else None

  @property
  def active(self):
    '''
    Return True if flow control turned off, False if flow control is on.
    '''
    return self._active

  def set_flow_cb(self, cb):
    '''
    Set a callback that will be called when the state of flow control has changed.
    The caller should use closures if they need to receive a handle to the channel
    on which flow control changes.
    '''
    self._flow_control_cb = cb
  
  def open(self):
    '''
    Open the channel for communication.
    '''
    args = Writer()
    args.write_shortstr('')
    self.send_frame( MethodFrame(self.channel_id, 20, 10, args) )
    self.channel.add_synchronous_cb( self._recv_open_ok )

  def _recv_open_ok(self, method_frame):
    pass

  def activate(self):
    '''
    Activate this channel (disable flow control).
    '''
    if not self._active:
      self._send_flow( True )

  def deactivate(self):
    '''
    Deactivate this channel (enable flow control).
    '''
    if self._active:
      self._send_flow( False )

  def _send_flow(self, active):
    '''
    Send a flow control command.
    '''
    args = Writer()
    args.write_bit( active )
    self.send_frame( MethodFrame(self.channel_id, 20, 20, args) )
    self.channel.add_synchronous_cb( self._recv_flow_ok )

  def _recv_flow(self, method_frame):
    '''
    Receive a flow control command from the broker
    '''
    self._active = method_frame.args.read_bit()
    
    args = Writer()
    args.write_bit( self._active )
    self.send_frame( MethodFrame(self.channel_id, 20, 21, args) )

    if self._flow_control_cb is not None:
      self._flow_control_cb()

  def _recv_flow_ok(self, method_frame):
    '''
    Receive a flow control ack from the broker.
    '''
    self._active = method_frame.args.read_bit()
    if self._flow_control_cb is not None:
      self._flow_control_cb()

  def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
    '''
    Close this channel.  Caller has the option of specifying the reason for
    closure and the class and method ids of the current frame in which an error
    occurred.  If in the event of an exception, the channel will be marked
    as immediately closed.  If channel is already closed, call is ignored.
    '''
    if self._closed: return

    self._close_info = {
      'reply_code'    : reply_code,
      'reply_text'    : reply_text,
      'class_id'      : class_id,
      'method_id'     : method_id
    }

    # exception likely due to race condition as connection is closing
    try:
      args = Writer()
      args.write_short( reply_code )
      args.write_shortstr( reply_text )
      args.write_short( class_id )
      args.write_short( method_id )
      self.send_frame( MethodFrame(self.channel_id, 20, 40, args) )
      
      self.channel.add_synchronous_cb( self._recv_close_ok )
    except:
      self.logger.error("Failed to close channel %d", 
        self.channel_id, exc_info=True)

    # Immediately set the closed flag so that no more frames can be sent
    self._closed = True

  def _recv_close(self, method_frame):
    '''
    Receive a close command from the broker.
    '''
    self._close_info = {
      'reply_code'    : method_frame.args.read_short(),
      'reply_text'    : method_frame.args.read_shortstr(),
      'class_id'      : method_frame.args.read_short(),
      'method_id'     : method_frame.args.read_short()
    }

    self._closed = True
    self.channel._closed_cb( final_frame=MethodFrame(self.channel_id, 20, 41) )

  def _recv_close_ok(self, method_frame):
    '''
    Receive a close ack from the broker.
    '''
    self._closed = True
    self.channel._closed_cb()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.classes import channel_class, ProtocolClass, ChannelClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

class ChannelClassTest(Chai):

  def setUp(self):
    super(ChannelClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = ChannelClass( ch )

  def test_init(self):
    expect(ProtocolClass.__init__).args('foo', a='b' )
    
    klass = ChannelClass.__new__(ChannelClass)
    klass.__init__('foo', a='b')

    assert_equals( 
      {
        11 : klass._recv_open_ok,
        20 : klass._recv_flow,
        21 : klass._recv_flow_ok,
        40 : klass._recv_close,
        41 : klass._recv_close_ok,
      }, klass.dispatch_map )
    assert_false( klass._closed )
    assert_equals(
      {
        'reply_code'    : 0,
        'reply_text'    : 'first connect',
        'class_id'      : 0,
        'method_id'     : 0
      }, klass._close_info )
    assert_true( klass._active )
    assert_equals( None, klass._flow_control_cb )

  def test_cleanup(self):
    self.klass._cleanup()
    assert_equals( None, self.klass._channel )
    assert_equals( None, self.klass.dispatch_map )

  def test_properties(self):
    self.klass._closed = 'yes'
    self.klass._close_info = 'ithappened'
    self.klass._active = 'record'
    assert_equals( 'yes', self.klass.closed )
    assert_equals( 'ithappened', self.klass.close_info )
    assert_equals( 'record', self.klass.active )

    self.klass._closed = False
    assert_equals( None, self.klass.close_info )

  def test_set_flow_cb(self):
    assert_equals( None, self.klass._flow_control_cb )
    self.klass.set_flow_cb( 'foo' )
    assert_equals( 'foo', self.klass._flow_control_cb )

  def test_open(self):
    writer = mock()
    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_shortstr ).args( '' )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 10, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_open_ok )

    self.klass.open()

  def test_recv_open_ok(self):
    self.klass._recv_open_ok('methodframe')

  def test_activate_when_not_active(self):
    self.klass._active = False
    expect( self.klass._send_flow ).args( True )
    self.klass.activate()

  def test_activate_when_active(self):
    self.klass._active = True
    stub( self.klass._send_flow )
    self.klass.activate()

  def test_deactivate_when_not_active(self):
    self.klass._active = False
    stub( self.klass._send_flow )
    self.klass.deactivate()

  def test_deactivate_when_active(self):
    self.klass._active = True
    expect( self.klass._send_flow ).args( False )
    self.klass.deactivate()

  def test_send_flow(self):
    writer = mock()
    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_bit ).args( 'active' )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 20, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_flow_ok )

    self.klass._send_flow('active')

  def test_recv_flow_no_cb(self):
    self.klass._flow_control_cb = None
    rframe = mock()
    writer = mock()
    expect( rframe.args.read_bit ).returns( 'active' )

    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_bit ).args( 'active' )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 21, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass._recv_flow(rframe)
    assert_equals( 'active', self.klass._active )

  def test_recv_flow_with_cb(self):
    self.klass._flow_control_cb = mock()
    rframe = mock()
    writer = mock()
    expect( rframe.args.read_bit ).returns( 'active' )

    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_bit ).args( 'active' )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 21, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass._flow_control_cb )

    self.klass._recv_flow(rframe)

  def test_recv_flow_ok_no_cb(self):
    self.klass._flow_control_cb = None
    rframe = mock()
    expect( rframe.args.read_bit ).returns( 'active' )

    self.klass._recv_flow_ok( rframe )
    assert_equals( 'active', self.klass._active )

  def test_recv_flow_ok_with_cb(self):
    self.klass._flow_control_cb = mock()
    rframe = mock()
    expect( rframe.args.read_bit ).returns( 'active' )
    expect( self.klass._flow_control_cb )

    self.klass._recv_flow_ok( rframe )
    assert_equals( 'active', self.klass._active )

  def test_close_when_not_closed(self):
    self.klass._closed = False
    writer = mock()
    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_short ).args( 'rcode' )
    expect( writer.write_shortstr ).args( 'reason' )
    expect( writer.write_short ).args( 'cid' )
    expect( writer.write_short ).args( 'mid' )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 40, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_close_ok )

    self.klass.close('rcode', 'reason', 'cid', 'mid')
    assert_true( self.klass._closed )
    assert_equals( {
        'reply_code'  : 'rcode',
        'reply_text'  : 'reason',
        'class_id'    : 'cid',
        'method_id'   : 'mid',
      }, self.klass._close_info )

  def test_close_when_closed(self):
    self.klass._closed = True
    stub( self.klass.send_frame )

    self.klass.close()

  def test_close_when_error_sending_frame(self):
    self.klass._closed = False
    writer = mock()
    expect( mock(channel_class, 'Writer') ).returns( writer )
    expect( writer.write_short ).args( 0 )
    expect( writer.write_shortstr ).args( '' )
    expect( writer.write_short ).args( 0 )
    expect( writer.write_short ).args( 0 )
    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 40, writer).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' ).raises( Exception('fail') )
    expect( self.klass.logger.error ).args( 
      'Failed to close channel %d', 42, exc_info=True )

    self.klass.close()
    assert_true( self.klass._closed )
    assert_equals( {
        'reply_code'  : 0,
        'reply_text'  : '',
        'class_id'    : 0,
        'method_id'   : 0,
      }, self.klass._close_info )

  def test_recv_close(self):
    rframe = mock()
    expect( rframe.args.read_short ).returns( 'rcode' )
    expect( rframe.args.read_shortstr ).returns( 'reason' )
    expect( rframe.args.read_short ).returns( 'cid' )
    expect( rframe.args.read_short ).returns( 'mid' )

    expect( mock(channel_class, 'MethodFrame') ).args(42, 20, 41).returns( 'frame' )
    expect( self.klass.channel._closed_cb ).args( final_frame='frame' )

    assert_false( self.klass._closed )
    self.klass._recv_close( rframe )
    assert_true( self.klass._closed )
    assert_equals( {
        'reply_code'  : 'rcode',
        'reply_text'  : 'reason',
        'class_id'    : 'cid',
        'method_id'   : 'mid',
      }, self.klass._close_info )

  def test_recv_close_ok(self):
    expect( self.klass.channel._closed_cb )

    self.klass._closed = False
    self.klass._recv_close_ok('frame')
    assert_true( self.klass._closed )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import struct
from collections import deque

from haigha.writer import Writer
from haigha.reader import Reader
from haigha.frames.frame import Frame

class HeaderFrame(Frame):
  '''
  Header frame for content.
  '''
  PROPERTIES = [
    ('content_type', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<15),
    ('content_encoding', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<14),
    ('application_headers', 'table', Reader.read_table, Writer.write_table, 1<<13),
    ('delivery_mode', 'octet', Reader.read_octet, Writer.write_octet, 1<<12),
    ('priority', 'octet', Reader.read_octet, Writer.write_octet, 1<<11),
    ('correlation_id', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<10),
    ('reply_to', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<9),
    ('expiration', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<8),
    ('message_id', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<7),
    ('timestamp', 'timestamp', Reader.read_timestamp, Writer.write_timestamp, 1<<6),
    ('type', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<5),
    ('user_id', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<4),
    ('app_id', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<3),
    ('cluster_id', 'shortstr', Reader.read_shortstr, Writer.write_shortstr, 1<<2)
  ]
  DEFAULT_PROPERTIES = True


  @classmethod
  def type(cls):
    return 2
  
  @property
  def class_id(self):
    return self._class_id
  
  @property
  def weight(self):
    return self._weight
  
  @property
  def size(self):
    return self._size

  @property
  def properties(self):
    return self._properties
  
  @classmethod
  def parse(self, channel_id, payload):
    '''
    Parse a header frame for a channel given a Reader payload.
    '''
    class_id = payload.read_short()
    weight = payload.read_short()
    size = payload.read_longlong()
    properties = {}

    # The AMQP spec is overly-complex when it comes to handling header frames.
    # The spec says that in addition to the first 16bit field, additional ones
    # can follow which /may/ then be in the property list (because bit flags
    # aren't in the list).  Properly implementing custom values requires the
    # ability change the properties and their types, which someone is welcome
    # to do, but seriously, what's the point? Because the complexity of parsing
    # and writing this frame directly impacts the speed at which messages can
    # be processed, there are two branches for both a fast parse which assumes
    # no changes to the properties and a slow parse. For now it's up to someone
    # using custom headers to flip the flag.
    if self.DEFAULT_PROPERTIES:
      flag_bits = payload.read_short()
      for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
        if flag_bits & mask:
          properties[ key ] = rfunc( payload )
    else:
      flags = []
      while True:
        flag_bits = payload.read_short()
        flags.append(flag_bits)
        if flag_bits & 1 == 0:
          break

      shift = 0
      for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
        if shift == 0:
          if not flags:
            break
          flag_bits, flags = flags[0], flags[1:]
          shift = 15
        if flag_bits & (1 << shift):
          properties[ key ] = rfunc( payload )
        shift -= 1

    return HeaderFrame( channel_id, class_id, weight, size, properties)
    
  def __init__(self, channel_id, class_id, weight, size, properties={}):
    Frame.__init__(self, channel_id)
    self._class_id = class_id
    self._weight = weight
    self._size = size
    self._properties = properties

  def __str__(self):
    return "%s[channel: %d, class_id: %d, weight: %d, size: %d, properties: %s]"%( self.__class__.__name__, self.channel_id, self._class_id, self._weight, self._size, self._properties )

  def write_frame(self, buf):
    '''
    Write the frame into an existing buffer.
    '''
    writer = Writer(buf)
    writer.write_octet( self.type() )
    writer.write_short( self.channel_id )
  
    # Track the position where we're going to write the total length
    # of the frame arguments.
    stream_args_len_pos = len(buf)
    writer.write_long(0)
    
    stream_method_pos = len(buf)

    writer.write_short( self._class_id )
    writer.write_short( self._weight )
    writer.write_longlong( self._size )

    # Like frame parsing, branch to faster code for default properties
    if self.DEFAULT_PROPERTIES:
      # Track the position where we're going to write the flags.
      flags_pos = len(buf)
      writer.write_short(0)
      flag_bits = 0
      for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
        val = self._properties.get(key, None)
        if val is not None:
          flag_bits |= mask
          wfunc(writer, val)
      writer.write_short_at( flag_bits, flags_pos )
    else:    
      shift = 15
      flag_bits = 0
      flags = []
      stack = deque()
      for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
        val = self._properties.get(key, None)
        if val is not None:
          if shift == 0:
            flags.append(flag_bits)
            flag_bits = 0
            shift = 15

          flag_bits |= (1 << shift)
          stack.append( (wfunc, val) )

        shift -= 1

      flags.append(flag_bits)
      for flag_bits in flags:
        writer.write_short(flag_bits)
      for method,val in stack:
        method(writer, val)

    # Write the total length back at the beginning of the frame
    stream_len = len(buf) - stream_method_pos
    writer.write_long_at( stream_len, stream_args_len_pos )

    writer.write_octet( 0xce )

HeaderFrame.register()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports import Transport

import gevent
from gevent.coros import Semaphore
from gevent import socket
from gevent import pool

#import socket

class GeventTransport(Transport):
  '''
  Transport using gevent backend. It relies on gevent's implementation of 
  sendall to send whole frames at a time. On the input side, it uses a gevent
  semaphore to ensure exclusive access to the socket and input buffer.

  NOTE:
  This is new to haigha and there may be integration issues with some versions
  of RabbitMQ. In particular, what may happen when a blocking call to send a
  frame allows another thread to queue another frame, but the frames can't be 
  interlaced due to how the protocol is defined or implemented. A lot of the
  'synchronous' calls have very specific expectations. If this becomes a
  problem then the quickest way to solve it would be to switch to non-blocking
  so that the same thread of execution is allowed to send all of its frames in
  sequence. In that case, the gevent implementation may be pushed into the
  EventSocket as another supported concurrency lib.

  Note also that the blocking nature of the sockets means that the any threads
  running IO should actively yield with a sleep(0) to ensure other threads are
  serviced. In a saturated environment, failure to do so will lead to 
  significant lags in signal handling or other IO. In practice, a typical
  client is attaching to other data stores so those could be enough to yield.
  '''

  def __init__(self, *args):
    super(GeventTransport,self).__init__(*args)

    self._buffer = bytearray()
    self._read_lock = Semaphore()
    self._write_lock = Semaphore()

  ###
  ### Transport API
  ###
  def connect(self, (host,port)):
    '''
    Connect assuming a host and port tuple.
    '''
    self._host = "%s:%s"%(host,port)
    self._sock = socket.socket()
    self._sock.setblocking( True )
    self._sock.settimeout( self.connection._connect_timeout )
    if self.connection._sock_opts:
      for k,v in self.connection._sock_opts.iteritems():
        family,type = k
        self._sock.setsockopt(family, type, v)
    self._sock.connect( (host,port) )

    # After connecting, switch to full-blocking mode.
    self._sock.settimeout( None )

  def read(self):
    '''
    Read from the transport. If no data is available, should return None.
    '''
    # NOTE: copying over this comment from Connection, because there is
    # knowledge captured here, even if the details are stale
    # Because of the timer callback to dataRead when we re-buffered, there's a
    # chance that in between we've lost the socket.  If that's the case, just
    # silently return as some code elsewhere would have already notified us.
    # That bug could be fixed by improving the message reading so that we consume
    # all possible messages and ensure that only a partial message was rebuffered,
    # so that we can rely on the next read event to read the subsequent message.
    if not hasattr(self,'_sock'):
      return None

    self._read_lock.acquire()
    try:
      data = self._sock.recv( self._sock.getsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF) )

      if len(data):
        if self.connection.debug > 1:
          self.connection.logger.debug( 'read %d bytes from %s'%(len(data), self._host) )
        if len(self._buffer):
          self._buffer.extend( data )
          data = self._buffer
          self._buffer = bytearray()
        return data
        
    finally:
      self._read_lock.release()

    self.connection.transport_closed( msg='error reading from %s'%(self._host) )

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''
    if not hasattr(self,'_sock'):
      return None

    self._read_lock.acquire()
    try:
      # data will always be a byte array
      if len(self._buffer):
        self._buffer.extend( data )
      else:
        self._buffer = bytearray(data)
    finally:
      self._read_lock.release()

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    if not hasattr(self,'_sock'):
      return None

    # MUST use a lock here else gevent could raise an exception if 2 greenlets
    # try to write at the same time. I was hoping that sendall() would do that
    # blocking for me, but I guess not.
    self._write_lock.acquire()
    try:
      self._sock.sendall( data )
    finally:
      self._write_lock.release()

    if self.connection.debug > 1:
      self.connection.logger.debug( 'sent %d bytes to %s'%(len(data), self._host) )
    
  def disconnect(self):
    '''
    Disconnect from the transport. Typically socket.close(). This call is 
    welcome to raise exceptions, which the Connection will catch.

    The transport is encouraged to allow for any pending writes to complete
    before closing the socket.
    '''
    if not hasattr(self,'_sock'):
      return None

    try:
      self._sock.close()
    finally:
      self._sock = None

class GeventPoolTransport(GeventTransport):

  def __init__(self, *args, **kwargs):
    super(GeventPoolTransport,self).__init__(*args)

    self._pool = kwargs.get('pool',None)
    if not self._pool:
      self._pool = gevent.pool.Pool()

  @property
  def pool(self):
    '''Get a handle to the gevent pool.'''
    return self._pool

  def process_channels(self, channels):
    '''
    Process a set of channels by calling Channel.process_frames() on each. 
    Some transports may choose to do this in unique ways, such as through 
    a pool of threads.

    The default implementation will simply iterate over them and call 
    process_frames() on each.
    '''
    for channel in channels:
      self._pool.spawn( channel.process_frames )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import struct
import sys
from collections import deque
from haigha.reader import Reader

class Frame(object):
  '''
  Base class for a frame.
  '''

  # Exceptions
  class FrameError(Exception): '''Base class for all frame errors'''
  class FormatError(FrameError): '''The frame was mal-formed.'''
  class InvalidFrameType(FrameError): '''The frame type is unknown.'''

  # Class data
  _frame_type_map = {}
  
  # Class methods
  @classmethod
  def register(cls):
    '''
    Register a frame type.
    '''
    cls._frame_type_map[ cls.type() ] = cls

  @classmethod
  def type(self):
    '''
    Fetch the type of this frame.  Should be an octet.
    '''
    raise NotImplementedError()

  @classmethod
  def read_frames(cls, reader):
    '''
    Read one or more frames from an IO stream.  Buffer must support file object
    interface.

    After reading, caller will need to check if there are bytes remaining in the
    stream.  If there are, then that implies that there is one or more incomplete
    frames and more data needs to be read.  The position of the cursor in the
    frame stream will mark the point at which the last good frame was read.  If
    the caller is expecting a sequence of frames and only received a part of that
    sequence, they are responsible for buffering those frames until the rest of
    the frames in the sequence have arrived.
    '''
    rval = deque()

    while True:
      frame_start_pos = reader.tell()
      try:
        frame = Frame._read_frame( reader )
      except Reader.BufferUnderflow:
        # No more data in the stream
        frame = None
      except Reader.ReaderError as e:
        # Some other format error
        raise Frame.FormatError, str(e), sys.exc_info()[-1]
      except struct.error as e:
        raise Frame.FormatError, str(e), sys.exc_info()[-1]

      if frame is None: 
        reader.seek( frame_start_pos )
        break

      rval.append( frame )

    return rval

  @classmethod
  def _read_frame(cls, reader):
    '''
    Read a single frame from a Reader.  Will return None if there is an incomplete
    frame in the stream.

    Raise MissingFooter if there's a problem reading the footer byte.
    '''
    frame_type = reader.read_octet()
    channel_id = reader.read_short()
    size = reader.read_long()
    
    payload = Reader(reader, reader.tell(), size)

    # Seek to end of payload
    reader.seek( size, 1 )

    ch = reader.read_octet()  # footer
    if ch != 0xce:
      raise Frame.FormatError(
        'Framing error, unexpected byte: %x.  frame type %x. channel %d, payload size %d',
        ch, frame_type, channel_id, size )

    frame_class = cls._frame_type_map.get( frame_type )
    if not frame_class:
      raise Frame.InvalidFrameType("Unknown frame type %x", frame_type)
    return frame_class.parse( channel_id, payload )


  # Instance methods
  def __init__(self, channel_id=-1):
    self._channel_id = channel_id

  @classmethod
  def parse(cls, channel_id, payload):
    '''
    Subclasses need to implement parsing of their frames.  Should return a new
    instance of their type.
    '''
    raise NotImplementedError()

  @property
  def channel_id(self):
    return self._channel_id

  def __str__(self):
    return "%s[channel: %d]"%( self.__class__.__name__, self.channel_id )

  def __repr__(self):
    # Have to actually call the method rather than __repr__==__str__ because
    # subclasses overload __str__
    return str(self)

  def write_frame(self, stream):
    '''
    Write this frame.
    '''
    raise NotImplementedError()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import gevent
from gevent.coros import Semaphore
from gevent import socket
from gevent.pool import Pool

from haigha.transports import *

class GeventTransportTest(Chai):

  def setUp(self):
    super(GeventTransportTest,self).setUp()

    self.connection = mock()
    self.transport = GeventTransport(self.connection)
    self.transport._host = 'server:1234'

  def test_init(self):
    assert_equals( bytearray(), self.transport._buffer )
    assert_true( isinstance(self.transport._read_lock,Semaphore) )
    assert_true( isinstance(self.transport._write_lock,Semaphore) )

  def test_connect(self):
    sock = mock()
    mock( gevent_transport, 'socket' )
    expect( gevent_transport.socket.socket ).returns( sock )
    self.connection._connect_timeout = 4.12
    self.connection._sock_opts = {
      ('family','tcp') : 34,
      ('range','ipv6') : 'hex'
    }

    expect( sock.setblocking ).args( True )
    expect( sock.settimeout ).args( 4.12 )
    expect( sock.setsockopt ).any_order().args( 'family', 'tcp', 34 ).any_order()
    expect( sock.setsockopt ).any_order().args( 'range', 'ipv6', 'hex' ).any_order()
    expect( sock.connect ).args( ('host',5309) )
    expect( sock.settimeout ).args( None )

    self.transport.connect( ('host',5309) )

  def test_read(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    self.transport.connection.debug = False

    expect( self.transport._read_lock.acquire )
    expect( self.transport._sock.getsockopt ).args(
      socket.SOL_SOCKET, socket.SO_RCVBUF ).returns( 4095 )
    expect( self.transport._sock.recv ).args(4095).returns('buffereddata')
    expect( self.transport._read_lock.release )

    assert_equals( 'buffereddata', self.transport.read() )

  def test_read_when_data_buffered(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    self.transport.connection.debug = False
    self.transport._buffer = bytearray('buffered')

    expect( self.transport._read_lock.acquire )
    expect( self.transport._sock.getsockopt ).any_args().returns( 4095 )
    expect( self.transport._sock.recv ).args(4095).returns('data')
    expect( self.transport._read_lock.release )

    assert_equals( 'buffereddata', self.transport.read() )
    assert_equals( bytearray(), self.transport._buffer )

  def test_read_when_debugging(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    self.transport.connection.debug = 2

    expect( self.transport._read_lock.acquire )
    expect( self.transport._sock.getsockopt ).any_args().returns( 4095 )
    expect( self.transport._sock.recv ).args(4095).returns('buffereddata')
    expect( self.transport.connection.logger.debug ).args(
      'read 12 bytes from server:1234' )
    expect( self.transport._read_lock.release )

    assert_equals( 'buffereddata', self.transport.read() )

  def test_read_when_socket_closes(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    self.transport.connection.debug = 2
    
    expect( self.transport._read_lock.acquire )
    expect( self.transport._sock.getsockopt ).any_args().returns( 4095 )
    expect( self.transport._sock.recv ).args(4095).returns('')
    expect( self.transport._read_lock.release )
    expect( self.transport.connection.transport_closed ).args(
      msg='error reading from server:1234' )
    
    self.transport.read()

  def test_read_when_no_sock(self):
    self.transport.read()

  def test_buffer(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    
    expect( self.transport._read_lock.acquire )
    expect( self.transport._read_lock.release )

    self.transport.buffer( bytearray('somedata') )
    assert_equals( bytearray('somedata'), self.transport._buffer )

  def test_buffer_when_already_buffered(self):
    self.transport._sock = mock()
    self.transport._read_lock = mock()
    self.transport._buffer = bytearray('some')
    
    expect( self.transport._read_lock.acquire )
    expect( self.transport._read_lock.release )

    self.transport.buffer( bytearray('data') )
    assert_equals( bytearray('somedata'), self.transport._buffer )

  def test_buffer_when_no_sock(self):
    self.transport.buffer('somedata')

  def test_write(self):
    self.transport._sock = mock()
    self.transport._write_lock = mock()
    self.transport.connection.debug = False

    expect( self.transport._write_lock.acquire )
    expect( self.transport._sock.sendall ).args( 'somedata' )
    expect( self.transport._write_lock.release )
    self.transport.write( 'somedata' )

  def test_write_when_sendall_fails(self):
    self.transport._sock = mock()
    self.transport._write_lock = mock()
    self.transport.connection.debug = False

    expect( self.transport._write_lock.acquire )
    expect( self.transport._sock.sendall ).args( 'somedata' ).raises(Exception('fail'))
    expect( self.transport._write_lock.release )
    assert_raises(Exception, self.transport.write, 'somedata' )

  def test_write_when_debugging(self):
    self.transport._sock = mock()
    self.transport.connection.debug = 2

    expect( self.transport._sock.sendall ).args( 'somedata' )
    expect( self.transport.connection.logger.debug ).args(
      'sent 8 bytes to server:1234' )

    self.transport.write( 'somedata' )

  def test_write_when_no_sock(self):
    self.transport.write('somedata')

  def test_disconnect(self):
    self.transport._sock = mock()
    expect( self.transport._sock.close )
    self.transport.disconnect()
    assert_equals( None, self.transport._sock )
  
  def test_disconnect_when_no_sock(self):
    self.transport.disconnect()

class GeventPoolTransportTest(Chai):

  def setUp(self):
    super(GeventPoolTransportTest,self).setUp()

    self.connection = mock()
    self.transport = GeventPoolTransport(self.connection)
    self.transport._host = 'server:1234'

  def test_init(self):
    assert_equals( bytearray(), self.transport._buffer )
    assert_true( isinstance(self.transport._read_lock,Semaphore) )
    assert_true( isinstance(self.transport.pool, Pool) )

    trans = GeventPoolTransport(self.connection, pool='inground')
    assert_equals( 'inground', trans.pool )

  def test_process_channels(self):
    chs = [ mock(), mock() ]
    self.transport._pool = mock()

    expect( self.transport._pool.spawn ).args( chs[0].process_frames )
    expect( self.transport._pool.spawn ).args( chs[1].process_frames )

    self.transport.process_channels( chs )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from frame import Frame
from method_frame import MethodFrame
from header_frame import HeaderFrame
from content_frame import ContentFrame
from heartbeat_frame import HeartbeatFrame



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.transports import *

class EventTransportTest(Chai):

  def setUp(self):
    super(EventTransportTest,self).setUp()

    self.connection = mock()
    self.transport = EventTransport(self.connection)
    self.transport._host = 'server'

  def test_sock_close_cb(self):
    expect( self.connection.transport_closed ).args( 
      msg='socket to server closed unexpectedly' )
    self.transport._sock_close_cb('sock')

  def test_sock_error_cb(self):
    expect( self.connection.transport_closed ).args( 
      msg='error on connection to server: amsg' )
    self.transport._sock_error_cb('sock', 'amsg')

  def test_sock_read_cb(self):
    expect( self.connection.read_frames )
    self.transport._sock_read_cb('sock')

  def test_connect(self):
    sock = mock()
    mock( event_transport, 'EventSocket' )
    self.connection._connect_timeout = 4.12
    self.connection._sock_opts = {
      ('family','tcp') : 34,
      ('range','ipv6') : 'hex'
    }

    expect( event_transport.EventSocket ).args( 
      read_cb = self.transport._sock_read_cb,
      close_cb = self.transport._sock_close_cb,
      error_cb = self.transport._sock_error_cb,
      debug = self.connection.debug,
      logger = self.connection.logger,
    ).returns( sock )
    expect( sock.setsockopt ).args( 'family', 'tcp', 34 ).any_order()
    expect( sock.setsockopt ).args( 'range', 'ipv6', 'hex' ).any_order()
    expect( sock.setblocking ).args( False )
    expect( sock.connect ).args( ('host',5309), timeout=4.12 )

    self.transport.connect( ('host',5309) )

  def test_read(self):
    self.transport._sock = mock()
    expect( self.transport._sock.read ).returns('buffereddata')
    assert_equals( 'buffereddata', self.transport.read() )

  def test_read_when_no_sock(self):
    self.transport.read()

  def test_buffer(self):
    self.transport._sock = mock()
    expect( self.transport._sock.buffer ).args( 'somedata' )
    self.transport.buffer( 'somedata' )

  def test_buffer_when_no_sock(self):
    self.transport.buffer('somedata')

  def test_write(self):
    self.transport._sock = mock()
    expect( self.transport._sock.write ).args( 'somedata' )
    self.transport.write( 'somedata' )

  def test_write_when_no_sock(self):
    self.transport.write('somedata')

  def test_disconnect(self):
    self.transport._sock = mock()
    self.transport._sock.close_cb = 'cb'
    expect( self.transport._sock.close )
    self.transport.disconnect()
    assert_equals( None, self.transport._sock.close_cb )
  
  def test_disconnect_when_no_sock(self):
    self.transport.disconnect()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports import Transport

from eventsocket import EventSocket
import event

class EventTransport(Transport):
  '''
  Transport using libevent-based EventSocket.
  '''

  ###
  ### EventSocket callbacks
  ###
  def _sock_close_cb(self, sock):
    self._connection.transport_closed(
      msg='socket to %s closed unexpectedly'%(self._host),
    )

  def _sock_error_cb(self, sock, msg, exception=None):
    self._connection.transport_closed(
      msg='error on connection to %s: %s'%(self._host, msg)
    )

  def _sock_read_cb(self, sock):
    self.connection.read_frames()
  
  ###
  ### Transport API
  ###
  def connect(self, (host,port)):
    '''
    Connect assuming a host and port tuple. Implemented as non-blocking, and
    will close the transport if there's an error
    '''
    self._host = "%s:%s"%(host,port)
    self._sock = EventSocket(
      read_cb=self._sock_read_cb,
      close_cb=self._sock_close_cb, 
      error_cb=self._sock_error_cb,
      debug=self.connection.debug, 
      logger=self.connection.logger )
    if self.connection._sock_opts:
      for k,v in self.connection._sock_opts.iteritems():
        family,type = k
        self._sock.setsockopt(family, type, v)
    self._sock.setblocking( False )
    self._sock.connect( (host,port), timeout=self.connection._connect_timeout )

  def read(self):
    '''
    Read from the transport. If no data is available, should return None.
    '''
    # NOTE: copying over this comment from Connection, because there is
    # knowledge captured here, even if the details are stale
    # Because of the timer callback to dataRead when we re-buffered, there's a
    # chance that in between we've lost the socket.  If that's the case, just
    # silently return as some code elsewhere would have already notified us.
    # That bug could be fixed by improving the message reading so that we consume
    # all possible messages and ensure that only a partial message was rebuffered,
    # so that we can rely on the next read event to read the subsequent message.
    if not hasattr(self,'_sock'):
      return None
    return self._sock.read()

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''
    if not hasattr(self,'_sock'):
      return None
    self._sock.buffer( data )

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    if not hasattr(self,'_sock'):
      return
    self._sock.write( data )
    
  def disconnect(self):
    '''
    Disconnect from the transport. Typically socket.close(). This call is 
    welcome to raise exceptions, which the Connection will catch.

    The transport is encouraged to allow for any pending writes to complete
    before closing the socket.
    '''
    if not hasattr(self,'_sock'):
      return

    # TODO: If there are bytes left on the output, queue the close for later.
    self._sock.close_cb = None
    self._sock.close()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.classes import transaction_class, ProtocolClass, TransactionClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

from collections import deque

class TransactionClassTest(Chai):

  def setUp(self):
    super(TransactionClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = TransactionClass( ch )

  def test_init(self):
    expect(ProtocolClass.__init__).args('foo', a='b' )
    
    klass = TransactionClass.__new__(TransactionClass)
    klass.__init__('foo', a='b')

    assert_equals( 
      {
        11 : klass._recv_select_ok,
        21 : klass._recv_commit_ok,
        31 : klass._recv_rollback_ok,
      }, klass.dispatch_map )
    assert_false( klass._enabled )
    assert_equals( deque(), klass._select_cb )
    assert_equals( deque(), klass._commit_cb )
    assert_equals( deque(), klass._rollback_cb )

  def test_cleanup(self):
    self.klass._cleanup()
    assert_equals( None, self.klass._select_cb )
    assert_equals( None, self.klass._commit_cb )
    assert_equals( None, self.klass._rollback_cb )
    assert_equals( None, self.klass._channel )
    assert_equals( None, self.klass.dispatch_map )

  def test_properties(self):
    self.klass._enabled = 'maybe'
    assert_equals( 'maybe', self.klass.enabled )

  def test_select_when_not_enabled_and_no_cb(self):
    self.klass._enabled = False
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 10).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_select_ok )

    self.klass.select()
    assert_true( self.klass.enabled )
    assert_equals( deque([None]), self.klass._select_cb )

  def test_select_when_not_enabled_with_cb(self):
    self.klass._enabled = False
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 10).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_select_ok )

    self.klass.select(cb='foo')
    assert_true( self.klass.enabled )
    assert_equals( deque(['foo']), self.klass._select_cb )

  def test_select_when_already_enabled(self):
    self.klass._enabled = True
    stub( self.klass.send_frame )

    assert_equals( deque(), self.klass._select_cb )
    self.klass.select()
    assert_equals( deque(), self.klass._select_cb )

  def test_recv_select_ok_with_cb(self):
    cb = mock()
    self.klass._select_cb.append( cb )
    self.klass._select_cb.append( mock() )
    expect(cb)
    self.klass._recv_select_ok( 'frame' )
    assert_equals( 1, len(self.klass._select_cb) )
    assert_false( cb in self.klass._select_cb )

  def test_recv_select_ok_without_cb(self):
    self.klass._select_cb.append( None )
    self.klass._select_cb.append( mock() )
    
    self.klass._recv_select_ok( 'frame' )
    assert_equals( 1, len(self.klass._select_cb) )
    assert_false( None in self.klass._select_cb )

  def test_commit_when_enabled_no_cb(self):
    self.klass._enabled = True
    
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 20).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_commit_ok )

    assert_equals( deque(), self.klass._commit_cb )
    self.klass.commit()
    assert_equals( deque([None]), self.klass._commit_cb )

  def test_commit_when_enabled_with_cb(self):
    self.klass._enabled = True
    
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 20).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_commit_ok )

    self.klass._commit_cb = deque(['blargh'])
    self.klass.commit(cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._commit_cb )

  def test_commit_raises_transactionsnotenabled_when_not_enabled(self):
    self.klass._enabled = False
    assert_raises( TransactionClass.TransactionsNotEnabled, self.klass.commit )

  def test_recv_commit_ok_with_cb(self):
    cb = mock()
    self.klass._commit_cb.append( cb )
    self.klass._commit_cb.append( mock() )
    expect(cb)
    
    self.klass._recv_commit_ok('frame')
    assert_equals( 1, len(self.klass._commit_cb) )
    assert_false( cb in self.klass._commit_cb )

  def test_recv_commit_ok_without_cb(self):
    self.klass._commit_cb.append( None )
    self.klass._commit_cb.append( mock() )
    
    self.klass._recv_commit_ok('frame')
    assert_equals( 1, len(self.klass._commit_cb) )
    assert_false( None in self.klass._commit_cb )

  def test_rollback_when_enabled_no_cb(self):
    self.klass._enabled = True
    
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 30).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_rollback_ok )

    assert_equals( deque(), self.klass._rollback_cb )
    self.klass.rollback()
    assert_equals( deque([None]), self.klass._rollback_cb )

  def test_rollback_when_enabled_with_cb(self):
    self.klass._enabled = True
    
    expect( mock(transaction_class, 'MethodFrame') ).args(42, 90, 30).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_rollback_ok )

    self.klass._rollback_cb = deque(['blargh'])
    self.klass.rollback(cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._rollback_cb )

  def test_rollback_raises_transactionsnotenabled_when_not_enabled(self):
    self.klass._enabled = False
    assert_raises( TransactionClass.TransactionsNotEnabled, self.klass.rollback )

  def test_recv_rollback_ok_with_cb(self):
    cb = mock()
    self.klass._rollback_cb.append( cb )
    self.klass._rollback_cb.append( mock() )
    expect(cb)
    
    self.klass._recv_rollback_ok('frame')
    assert_equals( 1, len(self.klass._rollback_cb) )
    assert_false( cb in self.klass._rollback_cb )

  def test_recv_rollback_ok_without_cb(self):
    self.klass._rollback_cb.append( None )
    self.klass._rollback_cb.append( mock() )
    
    self.klass._recv_rollback_ok('frame')
    assert_equals( 1, len(self.klass._rollback_cb) )
    assert_false( None in self.klass._rollback_cb )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames.frame import Frame
from haigha.reader import Reader
from haigha.writer import Writer

class MethodFrame(Frame):
  '''
  Frame which carries identifier for methods.
  '''

  @classmethod
  def type(cls):
    return 1
  
  @property
  def class_id(self):
    return self._class_id
  
  @property
  def method_id(self):
    return self._method_id
  
  @property
  def args(self):
    return self._args

  @classmethod
  def parse(self, channel_id, payload):
    class_id = payload.read_short()
    method_id = payload.read_short()
    return MethodFrame( channel_id, class_id, method_id, payload )

  def __init__(self, channel_id, class_id, method_id, args=None):
    Frame.__init__(self, channel_id)
    self._class_id = class_id
    self._method_id = method_id
    self._args = args

  def __str__(self):
    if isinstance(self.args, (Reader,Writer)):
      return "%s[channel: %d, class_id: %d, method_id: %d, args: %s]"%\
        (self.__class__.__name__, self.channel_id, self.class_id, self.method_id, str(self.args))
    else:
      return "%s[channel: %d, class_id: %d, method_id: %d, args: None]"%\
        ( self.__class__.__name__, self.channel_id, self.class_id, self.method_id)
  
  def write_frame(self, buf):
    writer = Writer(buf)
    writer.write_octet( self.type() )
    writer.write_short( self.channel_id )

    # Write a temporary value for the total length of the frame
    stream_args_len_pos = len(buf)
    writer.write_long(0)

    # Mark the point in the stream where we start writing arguments, *including*
    # the class and method ids.
    stream_method_pos = len(buf)

    writer.write_short(self.class_id)
    writer.write_short(self.method_id)

    # This is assuming that args is a Writer
    if self._args != None:
      writer.write( self._args.buffer() )

    # Write the total length back at the position we allocated
    stream_len = len(buf) - stream_method_pos
    writer.write_long_at( stream_len, stream_args_len_pos )

    # Write the footer
    writer.write_octet(0xce)
    
  
MethodFrame.register()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from transport import Transport
from event_transport import EventTransport
from gevent_transport import *



import haigha
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


requirements = map(str.strip, open('requirements.txt').readlines())

setup(
    name='haigha',
    version=haigha.VERSION,
    author='Vitaly Babiy, Aaron Westendorf',
    author_email="vbabiy@agoragames.com, aaron@agoragames.com",
    packages = ['haigha', 'haigha.frames', 'haigha.classes', 'haigha.transports'],
    install_requires = requirements,
    url='https://github.com/agoragames/haigha',
    license="LICENSE.txt",
    description='Event driven AMQP client library',
    long_description=open('README.rst').read(),
    keywords=['python', 'amqp', 'event', 'rabbitmq'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Libraries :: Python Modules",
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries'
    ]
)



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.classes import *
from haigha.frames import *
from haigha.exceptions import *

class Channel(object):
  '''
  Define a channel
  '''

  class InvalidClass(ChannelError): '''The method frame referenced an invalid class.  Non-fatal.'''
  class InvalidMethod(ChannelError): '''The method frame referenced an invalid method.  Non-fatal.'''
  class Inactive(ChannelError): '''Tried to send a content frame while the channel was inactive. Non-fatal.'''

  def __init__(self, connection, channel_id):
    '''Initialize with a handle to the connection and an id.'''
    self._connection = connection
    self._channel_id = channel_id
    
    self.channel = ChannelClass( self )
    self.exchange = ExchangeClass( self )
    self.queue = QueueClass( self )
    self.basic = BasicClass( self )
    self.tx = TransactionClass( self )

    self._class_map = {
      20 : self.channel,
      40 : self.exchange,
      50 : self.queue,
      60 : self.basic,
      90 : self.tx,
    }

    # Out-bound mix of pending frames and synchronous callbacks
    self._pending_events = deque()

    # Incoming frame buffer
    self._frame_buffer = deque()

    # Listeners for when channel closes
    self._close_listeners = set()

  @property
  def connection(self):
    return self._connection

  @property
  def channel_id(self):
    return self._channel_id

  @property
  def logger(self):
    return self._connection.logger

  @property
  def closed(self):
    return self.channel.closed

  @property
  def close_info(self):
    return self.channel.close_info

  @property
  def active(self):
    return self.channel.active

  def add_close_listener(self, listener):
    '''
    Add a listener for close events on this channel. The listener should be
    a callable that can take one argument, the channel that is closed. 
    Listeners will not be called in any particular order.
    '''
    self._close_listeners.add( listener )

  def remove_close_listener(self, listener):
    '''
    Remove a close event listener. Will do nothing if the listener is not
    registered.
    '''
    self._close_listeners.discard( listener )

  def open(self):
    '''
    Open this channel.  Routes to channel.open.
    '''
    self.channel.open()

  def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
    '''
    Close this channel.  Routes to channel.close.
    '''
    self.channel.close(reply_code, reply_text, class_id, method_id)

  def publish(self, *args, **kwargs):
    '''
    Standard publish.  See basic.publish.
    '''
    self.basic.publish( *args, **kwargs )

  def publish_synchronous(self, *args, **kwargs):
    '''
    Helper for publishing a message using transactions.  If 'cb' keyword arg
    is supplied, will be called when the transaction is committed.
    '''
    cb = kwargs.pop('cb', None)
    self.tx.select()
    self.basic.publish( *args, **kwargs )
    self.tx.commit( cb=cb )

  def dispatch(self, method_frame):
    '''
    Dispatch a method.
    '''
    klass = self._class_map.get( method_frame.class_id )
    if klass:
      klass.dispatch( method_frame )
    else:
      raise Channel.InvalidClass( "class %d is not supported on channel %d", 
        method_frame.class_id, self.channel_id )

  def buffer_frame(self, frame):
    '''
    Buffer an input frame.  Will append to current list of frames and ensure
    there's a pending event to process the queue.
    '''
    self._frame_buffer.append( frame )

  def process_frames(self):
    '''
    Process the input buffer.
    '''
    while len(self._frame_buffer):
      try:
        # It would make sense to call next_frame, but it's technically faster
        # to repeat the code here.
        frame = self._frame_buffer.popleft()
        self.dispatch( frame )
      except ProtocolClass.FrameUnderflow:
        return
      except:
        self.logger.error( 
          "Failed to dispatch %s", frame, exc_info=True )
        self.close( 500, "Failed to dispatch %s"%(str(frame)) )
        return

  def next_frame(self):
    '''
    Pop the next frame off the input queue. If the queue is empty, will return
    None.
    '''
    if len( self._frame_buffer ):
      return self._frame_buffer.popleft()
    return None

  def requeue_frames(self, frames):
    '''
    Requeue a list of frames. Will append to the head of the frame buffer.
    Frames should be in reverse order. Really only used to support BasicClass
    content consumers
    '''
    self._frame_buffer.extendleft( frames )

  def send_frame(self, frame):
    '''
    Queue a frame for sending.  Will send immediately if there are no pending
    synchronous transactions on this connection.
    '''
    if self.closed:
      if self.close_info and len(self.close_info['reply_text'])>0:
        raise ChannelClosed(
          "channel %d is closed: %s : %s",
          self.channel_id,
          self.close_info['reply_code'],
          self.close_info['reply_text'] )
      raise ChannelClosed()


    # If there's any pending event at all, then it means that when the current
    # dispatch loop started, all possible frames were flushed and the remaining
    # item(s) starts with a sync callback.  After careful consideration, it
    # seems that it's safe to assume the len>0 means to buffer the frame. The
    # other advantage here is 
    if not len(self._pending_events):
      if not self.channel.active and isinstance( frame, (ContentFrame,HeaderFrame) ):
        raise Channel.Inactive( "Channel %d flow control activated", self.channel_id )
      self._connection.send_frame(frame)
    else:
      self._pending_events.append( frame )

  def add_synchronous_cb(self, cb):
    '''
    Add an expectation of a callback to release a synchronous transaction.
    '''
    self._pending_events.append( cb )

  def clear_synchronous_cb(self, cb):
    '''
    If the callback is the current expected callback, will clear it off the
    stack.  Else will raise in exception if there's an expectation but this
    doesn't satisfy it.
    '''
    if len(self._pending_events):
      ev = self._pending_events[0]

      # We can't have a strict check using this simple mechanism, because we
      # could be waiting for a synch response while messages are being published.
      # So for now, if it's not in the list, do a check to see if the callback
      # is in the pending list, and if so, then raise, because it means we
      # received stuff out of order.  Else just pass it through.
      # Note that this situation could happen on any broker-initiated message.
      if ev==cb:
        self._pending_events.popleft()
        self._flush_pending_events()
      elif cb in self._pending_events:
        raise ChannelError("Expected synchronous callback %s, got %s", ev, cb)

  def _flush_pending_events(self):
    '''
    Send pending frames that are in the event queue.
    '''
    while len(self._pending_events) and isinstance(self._pending_events[0],Frame):
      self._connection.send_frame( self._pending_events.popleft() )

  def _closed_cb(self, final_frame=None):
    '''
    "Private" callback from the ChannelClass when a channel is closed. Only
    called after broker initiated close, or we receive a close_ok. Caller has
    the option to send a final frame, to be used to bypass any synchronous or
    otherwise-pending frames so that the channel can be cleanly closed.
    '''
    # delete all pending data and send final frame if thre is one. note that
    # it bypasses send_frame so that even if the closed state is set, the frame
    # is published.
    if final_frame:
      self._connection.send_frame( final_frame )
    self._pending_events = deque()
    self._frame_buffer = deque()

    # clear out other references for faster cleanup
    for protocol_class in self._class_map.values():
      protocol_class._cleanup()
    self._connection = None
    self.channel = None
    self.exchange = None
    self.queue = None
    self.basic = None
    self.tx = None
    self._class_map = None

    for listener in self._close_listeners:
      listener( self )
    self._close_listeners = set()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import logging
from chai import Chai

from haigha import connection, VERSION
from haigha.connection import Connection

class ConnectionTest(Chai):
  
  def setUp(self):
    super(ConnectionTest,self).setUp()

    self.connection = Connection.__new__( Connection )
    self.connection._debug = False
    self.connection._logger = self.mock()
    self.connection._user = 'guest'
    self.connection._password = 'guest'
    self.connection._host = 'localhost'
    self.connection._vhost = '/'
    self.connection._connect_timeout = 5
    self.connection._sock_opts = None
    self.connection._sock = None # mock anything?
    self.connection._heartbeat = None
    self.connection._reconnect_cb = self.mock()
    self.connection._close_cb = self.mock()
    self.connection._login_method = 'AMQPLAIN'
    self.connection._locale = 'en_US'
    self.connection._client_properties = None
    self.connection._properties = {
      'library': 'Haigha',
      'library_version': 'x.y.z',
    }
    self.connection._closed = False
    self.connection._connected = False
    self.connection._close_info = {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    }
    self.connection._channels = {
      0 : self.mock()
    }
    self.connection._login_response = 'loginresponse'
    self.connection._channel_counter = 0
    self.connection._channel_max = 65535
    self.connection._frame_max = 65535
    self.connection._frames_read = 0
    self.connection._frames_written = 0
    self.connection._strategy = self.mock()
    self.connection._output_frame_buffer = []
    self.connection._transport = mock()

  def test_init_without_keyword_args(self):
    conn = Connection.__new__( Connection )
    strategy = mock()
    mock( connection, 'ConnectionChannel' )

    expect(connection.ConnectionChannel).args( conn, 0 ).returns( 'connection_channel' )
    expect(conn.connect).args( 'localhost', 5672 )

    conn.__init__()
    
    self.assertFalse( conn._debug )
    self.assertEqual( logging.root, conn._logger )
    self.assertEqual( 'guest', conn._user )
    self.assertEqual( 'guest', conn._password )
    self.assertEqual( 'localhost', conn._host )
    self.assertEqual( 5672, conn._port )
    self.assertEqual( '/', conn._vhost )
    self.assertEqual( 5, conn._connect_timeout )
    self.assertEqual( None, conn._sock_opts )
    self.assertEqual( None, conn._sock )
    self.assertEqual( None, conn._heartbeat )
    self.assertEqual( None, conn._reconnect_cb )
    self.assertEqual( None, conn._close_cb )
    self.assertEqual( 'AMQPLAIN', conn._login_method )
    self.assertEqual( 'en_US', conn._locale )
    self.assertEqual( None, conn._client_properties )
    self.assertEqual( conn._properties, {
      'library': 'Haigha',
      'library_version': VERSION,
    } )
    self.assertFalse( conn._closed )
    self.assertFalse( conn._connected )
    self.assertEqual( conn._close_info, {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    } )
    self.assertEqual( {0:'connection_channel'}, conn._channels )
    self.assertEqual( '\x05LOGINS\x00\x00\x00\x05guest\x08PASSWORDS\x00\x00\x00\x05guest', conn._login_response )
    self.assertEqual( 0, conn._channel_counter )
    self.assertEqual( 65535, conn._channel_max )
    self.assertEqual( 65535, conn._frame_max )
    self.assertEqual( [], conn._output_frame_buffer )
    self.assertTrue( isinstance(conn._transport, connection.EventTransport) )

  def test_properties(self):
    self.assertEqual( self.connection._logger, self.connection.logger )
    self.assertEqual( self.connection._debug, self.connection.debug )
    self.assertEqual( self.connection._frame_max, self.connection.frame_max )
    self.assertEqual( self.connection._channel_max, self.connection.channel_max )
    self.assertEqual( self.connection._frames_read, self.connection.frames_read )
    self.assertEqual( self.connection._frames_written, self.connection.frames_written )

  def test_connect(self):
    self.connection._connected = 'maybe'
    self.connection._closed = 'possibly'
    self.connection._debug = 'sure'
    self.connection._connect_timeout = 42
    self.connection._sock_opts = {
      ('f1','t1') : 5,
      ('f2','t2') : 6
    }

    expect( self.connection._transport.connect ).args( ('host',5672) )
    expect( self.connection._transport.write ).args( 'AMQP\x00\x00\x09\x01' )

    self.connection.connect( 'host', 5672 )
    assert_false( self.connection._connected )
    assert_false( self.connection._closed )
    assert_equals( self.connection._close_info,
      {
      'reply_code'    : 0,
      'reply_text'    : 'failed to connect to host:5672',
      'class_id'      : 0,
      'method_id'     : 0
      } )
    assert_equals( 'host:5672', self.connection._host )

  def test_disconnect_when_transport_disconnects(self):
    self.connection._connected = 'yup'

    expect( self.connection._transport.disconnect )
    self.connection.disconnect()

    assert_false( self.connection._connected )
    assert_equals( None, self.connection._transport )

  def test_disconnect_when_transport_disconnects_with_error(self):
    self.connection._connected = 'yup'
    self.connection._host = 'server'

    expect( self.connection._transport.disconnect ).raises( Exception('fail') )
    expect( self.connection.logger.error ).args( "Failed to disconnect from %s", 'server', exc_info=True )
    self.connection.disconnect()

    assert_false( self.connection._connected )
    assert_equals( None, self.connection._transport )

  def test_transport_closed_with_no_args(self):
    self.connection._host = 'server'
    self.connection._connected = 'yes'

    expect( self.connection.logger.warning ).args( 'transport to server closed : unknown cause' ) 
    expect( self.connection._callback_close )

    self.connection.transport_closed()

    assert_equals( 0, self.connection._close_info['reply_code'] )
    assert_equals( 'transport to server closed : unknown cause', self.connection._close_info['reply_text'] )
    assert_equals( 0, self.connection._close_info['class_id'] )
    assert_equals( 0, self.connection._close_info['method_id'] )

  def test_next_channel_id_when_less_than_max(self):
    self.connection._channel_counter = 32
    self.connection._channel_max = 23423
    assert_equals( 33, self.connection._next_channel_id() )
  
  def test_next_channel_id_when_at_max(self):
    self.connection._channel_counter = 32
    self.connection._channel_max = 32
    assert_equals( 1, self.connection._next_channel_id() )

  def test_channel_creates_new_when_not_at_limit(self):
    ch = mock()
    expect( self.connection._next_channel_id ).returns( 1 )
    mock( connection, 'Channel' )
    expect( connection.Channel ).args( self.connection, 1).returns( ch )
    expect( ch.add_close_listener ).args( self.connection._channel_closed )
    expect( ch.open )

    self.assert_equals( ch, self.connection.channel() )
    self.assert_equals( ch, self.connection._channels[1] )

  def test_channel_finds_the_first_free_channel_id(self):
    self.connection._channels[1] = 'foo'
    self.connection._channels[2] = 'bar'
    self.connection._channels[4] = 'cat'
    ch = mock()
    expect( self.connection._next_channel_id ).returns( 1 )
    expect( self.connection._next_channel_id ).returns( 2 )
    expect( self.connection._next_channel_id ).returns( 3 )
    mock( connection, 'Channel' )
    expect( connection.Channel ).args( self.connection, 3 ).returns( ch )
    expect( ch.add_close_listener ).args( self.connection._channel_closed )
    expect( ch.open )

    self.assert_equals( ch, self.connection.channel() )
    self.assert_equals( ch, self.connection._channels[3] )

  def test_channel_raises_toomanychannels(self):
    self.connection._channels[1] = 'foo'
    self.connection._channels[2] = 'bar'
    self.connection._channels[4] = 'cat'
    self.connection._channel_max = 3
    assert_raises( Connection.TooManyChannels, self.connection.channel )

  def test_channel_returns_cached_instance_if_known(self):
    self.connection._channels[1] = 'foo'
    assert_equals( 'foo', self.connection.channel(1) )

  def test_channel_raises_invalidchannel_if_unknown_id(self):
    assert_raises( Connection.InvalidChannel, self.connection.channel, 42 )

  def test_channel_closed(self):
    ch = mock()
    ch.channel_id = 42
    self.connection._channels[42] = ch

    self.connection._channel_closed(ch)
    assert_false( 42 in self.connection._channels )

    ch.channel_id = 500424834
    self.connection._channel_closed(ch)

  def test_close(self):
    self.connection._channels[0] = mock()
    expect( self.connection._channels[0].close ).times(2)
    
    self.connection.close()
    assert_equals( {'reply_code':0, 'reply_text':'', 'class_id':0, 'method_id':0}, 
      self.connection._close_info )
    
    self.connection.close(1, 'foo', 2, 3)
    assert_equals( {'reply_code':1, 'reply_text':'foo', 'class_id':2, 'method_id':3}, 
      self.connection._close_info )

  def test_callback_close_when_no_cb(self):
    self.connection._close_cb = None
    self.connection._callback_close()

  def test_callback_close_when_user_cb(self):
    self.connection._close_cb = mock()
    expect( self.connection._close_cb )
    self.connection._callback_close()

  def test_callback_close_raises_sysexit_when_user_cb_does(self):
    self.connection._close_cb = mock()
    expect( self.connection._close_cb ).raises( SystemExit() )
    assert_raises( SystemExit, self.connection._callback_close )

  def test_callback_close_logs_when_user_cb_fails(self):
    self.connection._close_cb = mock()
    expect( self.connection._close_cb ).raises( 'fail!' )
    expect( self.connection.logger.error ).args( str )
    self.connection._callback_close()

  def test_read_frames_when_no_transport(self):
    self.connection._transport = None
    self.connection.read_frames()
    assert_equals( 0, self.connection._frames_read )

  def test_read_frames_when_transport_returns_no_data(self):
    expect( self.connection._transport.read ).returns( None )
    self.connection.read_frames()
    assert_equals( 0, self.connection._frames_read )

  def test_read_frames_when_transport_when_frame_data_and_no_debug_and_no_buffer(self):
    reader = mock()
    frame = mock()
    frame.channel_id = 42
    channel = mock()
    mock( connection, 'Reader' )
    
    expect( self.connection._transport.read ).returns( 'data' )
    expect( connection.Reader ).args( 'data' ).returns( reader )
    expect( connection.Frame.read_frames ).args( reader ).returns( [frame] )
    expect( self.connection.channel ).args( 42 ).returns( channel )
    expect( channel.buffer_frame ).args( frame )
    expect( self.connection._transport.process_channels ).args( set([channel]) )
    expect( reader.tell ).returns( 4 )

    self.connection.read_frames()
    assert_equals( 1, self.connection._frames_read )

  def test_read_frames_when_transport_when_frame_data_and_debug_and_buffer(self):
    reader = mock()
    frame = mock()
    frame.channel_id = 42
    channel = mock()
    mock( connection, 'Reader' )
    self.connection._debug = 2
    
    expect( self.connection._transport.read ).returns( 'data' )
    expect( connection.Reader ).args( 'data' ).returns( reader )
    expect( connection.Frame.read_frames ).args( reader ).returns( [frame] )
    expect( self.connection.logger.debug ).args( 'READ: %s', frame )
    expect( self.connection.channel ).args( 42 ).returns( channel )
    expect( channel.buffer_frame ).args( frame )
    expect( self.connection._transport.process_channels ).args( set([channel]) )
    expect( reader.tell ).times(2).returns( 2 )
    expect( self.connection._transport.buffer ).args( 'ta' )

    self.connection.read_frames()
    assert_equals( 1, self.connection._frames_read )

  def test_flush_buffered_frames(self):
    self.connection._output_frame_buffer = ['frame1', 'frame2']
    expect( self.connection.send_frame ).args( 'frame1' )
    expect( self.connection.send_frame ).args( 'frame2' )

    self.connection._flush_buffered_frames()
    assert_equals( [], self.connection._output_frame_buffer )

  def test_send_frame_when_connected_and_transport_and_no_debug(self):
    frame = mock()
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )
    
    self.connection._connected = True
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_not_connected_and_not_channel_0(self):
    frame = mock()
    frame.channel_id = 42
    stub( frame.write_frame )
    stub( self.connection._transport.write )
    
    self.connection._connected = False
    self.connection.send_frame( frame )
    assert_equals( [frame], self.connection._output_frame_buffer )

  def test_send_frame_when_not_connected_and_channel_0(self):
    frame = mock()
    frame.channel_id = 0
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )
    
    self.connection._connected = False
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_debugging(self):
    frame = mock()
    expect( self.connection.logger.debug ).args( 'WRITE: %s', frame )
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )
    
    self.connection._connected = True
    self.connection._debug = 2
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_closed(self):
    self.connection._closed = True
    self.connection._close_info['reply_text'] = 'failed'
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )
    
    self.connection._close_info['reply_text'] = ''
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )
    
    self.connection._close_info = None
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.message import Message


class MessageTest(Chai):

  def test_init_no_args(self):
    m = Message()
    self.assertEquals( None, m._body )
    self.assertEquals( None, m._delivery_info )
    self.assertEquals( {}, m._properties )

  def test_init_with_args(self):
    m = Message( 'foo', 'delivery', foo='bar' )
    self.assertEquals( 'foo', m._body )
    self.assertEquals( 'delivery', m._delivery_info )
    self.assertEquals( {'foo':'bar'}, m._properties )

    m = Message( u'D\xfcsseldorf' )
    self.assertEquals( 'D\xc3\xbcsseldorf', m._body )
    self.assertEquals( {'content_encoding':'utf-8'}, m._properties )

  def test_properties(self):
    m = Message( 'foo', 'delivery', foo='bar' )
    self.assertEquals( 'foo', m.body )
    self.assertEquals( 'delivery', m.delivery_info )
    self.assertEquals( {'foo':'bar'}, m.properties )

  def test_len(self):
    m = Message('foobar')
    self.assertEquals( 6, len(m) )

  def test_nonzero(self):
    m = Message()
    self.assertTrue( m )

  def test_eq(self):
    l=Message(); r=Message()
    self.assertEquals( l, r )

    l=Message('foo'); r=Message('foo')
    self.assertEquals( l, r )

    l=Message(foo='bar'); r=Message(foo='bar')
    self.assertEquals( l, r )

    l=Message('hello', foo='bar'); r=Message('hello', foo='bar')
    self.assertEquals( l, r )

    l=Message('foo'); r=Message('bar')
    self.assertNotEquals( l, r )

    l=Message(foo='bar'); r=Message(foo='brah')
    self.assertNotEquals( l, r )

    l=Message('hello', foo='bar'); r=Message('goodbye', foo='bar')
    self.assertNotEquals( l, r )

    l=Message('hello', foo='bar'); r=Message('hello', foo='brah')
    self.assertNotEquals( l, r )

    self.assertNotEquals( Message(), object() )

  def test_str(self):
    m = Message( 'foo', 'delivery', foo='bar' )
    str(m)
